#!/usr/bin/env python

"""Tests for `phase_curve_tools` package."""

import pytest


from phase_curve_tools import bowellCalcAbsMag, calcHG1G2



def test_hg_simple():
    assert bowellCalcAbsMag(18.168, 1.997, 1.187, 22.9769, .15) == pytest.approx(15.201,.05)

def test_hg():
    assert bowellCalcAbsMag(18.168, 1.997, 1.187, 22.9769, .15, simple=False) == pytest.approx(15.201,.05)


def test_hg1g2():

    assert calcHG1G2(18.168, 1.997, 1.187, 22.9769, .1, .2) == pytest.approx(15.201,.05)


"""Top-level package for Phase Curve Tools."""

__author__ = """Daniel Kramer"""
__email__ = 'drk98@nau.edu'
__version__ = '0.4.0'


from .hg import bowellCalcAbsMag, calcPhaseCurve
from .reducedMag import calcReducedMag
from .hg1g2 import calcHG1G2

from .useAstroQuery import HorAbsMags
from .bestFit import BestFit, HGResult


cimport numpy as np
import numpy as np
cimport cython


ctypedef fused valid_arg:
    np.ndarray
    cython.float
    cython.double


cdef extern from "constants.h":
    cdef cython.float M_1_PI 
    cdef cython.float M_1_5PI 


cdef _calcPhis( valid_arg phaseAngle):
    cdef valid_arg phi1, phi2, phi3, pa_rad
    pa_rad = phaseAngle * np.pi/180
    phi1 = 1- (6 * M_1_PI * pa_rad)
    phi2 = 1- (9 * M_1_5PI * pa_rad)

    phi3 = np.exp(-4*np.pi*np.power(np.tan(.5 * pa_rad), 2/3))

    return phi1, phi2, phi3

cdef _calcPhaseInt( valid_arg phaseAngle,  float G1,  float G2):
    cdef valid_arg phi1, phi2, phi3

    phi1, phi2, phi3 = _calcPhis(phaseAngle)

    return G1*phi1 + G2*phi2 + (1-G1-G2)*phi3

    
cdef _HG1G2( valid_arg apmag,  valid_arg helioDist,  valid_arg obsDist,  valid_arg phaseAngle , cython.float G1,  cython.float G2):
    cdef valid_arg phaseInt

    phaseInt = _calcPhaseInt(phaseAngle, G1, G2)

    return apmag - (5 * np.log10(helioDist*obsDist) - 2.5*np.log10(phaseInt))

def calcHG1G2( valid_arg apmag,  valid_arg helioDist,  valid_arg obsDist,  valid_arg phaseAngle , cython.float G1,  cython.float G2):
    r"""Calculate the absolute mags for an object from the HG1G2 system\ :footcite:p:`MUINONEN2010542`
    
    

    :param apmag: The apparent magnitudes to use
    :type apmag: np.ndarray
    :param helioDist: The distance from the object to the Sun (AU)
    :type helioDist: np.ndarray
    :param obsDist: The distance from the object to the Observer/Earth (AU)
    :type obsDist: np.ndarray
    :param phaseAngle: The phase angle, the sun - object - observer angle (degrees)
    :type phaseAngle: np.ndarray
    :param G1: The G1 value
    :type G1: float
    :param G2: The G2 value
    :type G2: float
    :return: The absolute magnitudes for the observations.
    :rtype: np.ndarray
    """

    return _HG1G2(apmag, helioDist, obsDist, phaseAngle, G1, G2)


from dataclasses import dataclass
from scipy.optimize import minimize
import numpy as np
from typing import Tuple, Union
from astroquery.jplhorizons import Horizons
from .hg import calcPhaseCurve


@dataclass(frozen=True)
class HGResult:
    """Class for hoding the reult of HG minimumzation

    :param H: The Bowell H value
    :type H: float
    :param G: The Bowell G value
    :type G: float
    :param sigH: The error in the Bowell H value
    :type sigH: float
    :param sigG: The error in the Bowell G value
    :type sigG: float
    :param success: If the minimization succedded (according to scipy)
    :type success: bool
    :param atEdge: If the derived G value is at the edge of the porvided G bounds
    :type atEdge: bool
    :param n: The number of observations used
    :type n: int
    """

    H: float
    G: float
    sigH: float = 0
    sigG: float = 0
    success: bool = False
    atEdge: bool = False
    n: int = 0

    @property
    def fullSuccess(self) -> bool:
        """fullSuccess If the minimzation succeded and the G value is not at an edge

        :getter: Returns if there was a full succedd
        :type: bool
        """
        return self.success and not self.atEdge


class BestFit:
    """Class to find the best fit in the Bowell HG system.

    If the helioDist nor obsDist is provided, then Horizons will be queried, so the object ID and times are needed. They don't have to be provided if helioDist and obsDist are provided.

    If no phase angles are provided, then they will be computed, assuming the distance from the observer/Earth to the Sun is always 1 AU.

    :param mags: The magnitudes of the observations
    :type mags: np.ndarray
    :param errors: The errors of the magnitudes, defaults to None
    :type errors: np.ndarray, optional
    :param helioDist: The distance from the Sun to the object (AU), defaults to None
    :type helioDist: np.ndarray, optional
    :param obsDist: The distance from the Observer/Earth to the object (AU), defaults to None
    :type obsDist: np.ndarray, optional
    :param phaseAngle: The phase angle of the observations (deg), defaults to None
    :type phaseAngle: np.ndarray, optional
    :param times: The times of the observations (JD), defaults to None
    :type times: np.ndarray, optional
    :param objectID: The object ID, needs to be compatable with JPL Horizons, defaults to None
    :type objectID: Union[str, int], optional
    :raises ValueError: Needs the object ID and the times of the observations if helioDist or obsDist is not provided
    """

    def __init__(
        self,
        mags: np.ndarray,
        errors: np.ndarray = None,
        helioDist: np.ndarray = None,
        obsDist: np.ndarray = None,
        phaseAngle: np.ndarray = None,
        times: np.ndarray = None,
        objectID: Union[str, int] = None,
    ) -> None:
        """Constructor"""

        self.helioDist: np.ndarray = helioDist
        self.obsDist: np.ndarray = obsDist
        self.phaseAngle: np.ndarray = phaseAngle

        if helioDist is None or obsDist is None:

            # Need an object ID and times to query horizions
            if objectID is None:
                raise ValueError(
                    "No heliocentric distances orobserver-centric distances provided and no object id provided. Unable to procede."
                )
            if times is None:
                raise ValueError(
                    "No heliocentric distances orobserver-centric distances provided and no times provided. Unable to procede."
                )

            eph = Horizons(id=objectID, epochs=times).ephemerides()
            self.helioDist = self.eph["r"].data
            self.obsDist = self.eph["delta"].data
            self.phaseAngle = self.eph["alpha"].data
        elif phaseAngle is None:
            self.phaseAngle = BestFit.calcPhaseAngle(self.helioDist, self.obsDist)

        self.mags: np.ndarray = mags
        self.errors: np.ndarray = errors

        self.reducedMagHG: np.ndarray = None

    @staticmethod
    def calcPhaseAngle(
        helioDist: np.ndarray,
        obsDist: np.ndarray,
        sunObsDist: Union[np.ndarray, float] = 1,
    ) -> np.ndarray:
        r"""calcPhaseAngle Calculate the phase angles.
        To calculate the s-t-o phase angle, using the law of cosines:

        .. math:: \alpha=\cos^{-1}\left(\dfrac{\Delta^2+r^2-d^2}{2\Delta{r}}\right)

        where :math:`\Delta` is the heliocentric distance, :math:`r` is the observer/Earth distance, and :math:`d` is the distance from the observer/Earth to the Sun.

        :param helioDist: The heliocentric distance (AU)
        :type helioDist: np.ndarray
        :param obsDist: The observer distance (AU)
        :type obsDist: np.ndarray
        :param sunObsDist: The distance from the observer/Earth to the Sun (AU), defaults to 1
        :type sunObsDist: Union[np.ndarray, float], optional
        :return: The phase angles (deg)
        :rtype: np.ndarray
        """
        return np.arccos(
            (helioDist**2 + obsDist**2 + sunObsDist**2)
            / (2 * helioDist * obsDist)
        )

    def fitHG(
        self,
        H_0: float = None,
        G_0: float = 0.15,
        HRange: Tuple = (4.69236933828342, 29.276432904419803),
        GRange: Tuple = (-0.301744075798055, 0.9073912874142601),
        errorSimulation: int = 30,
    ) -> HGResult:
        r"""fitHG Fit the data to the Bowell HG system\ :footcite:p:`bowellHG`.

        If errors are provided, then :math:`\sigma_H,\sigma_G` will be generated with the reduced magnitudes being shifited with the normal distribution with :math:`\sigma` is the error of the observations.

        :param H_0: The initial H guess, if None, then will use the reduced magnitude at the minimum phase angle, defaults to None
        :type H_0: float, optional
        :param G_0: The inital G guess, defaults to .15
        :type G_0: float, optional
        :param HRange: The valid range for H values, defaults to (4.69236933828342, 29.276432904419803), which are :math:`\pm10\%` the values from \ :footcite:t:`verdes`
        :type HRange: Tuple, optional
        :param GRange: The valid range for G values, defaults to (-0.301744075798055, 0.9073912874142601), which are :math:`\pm10\%` the values from \ :footcite:t:`verdes`
        :type GRange: Tuple, optional
        :return: The result of the fit
        :rtype: HGResult
        """

        if self.reducedMag is None:
            self.reducedMagHG = self.mags - 5 * np.log10(self.helioDist * self.obsDist)

        if H_0 is None:
            H_0 = self.reducedMag[np.argmin(self.phaseAngle)]

        res = minimize(self._HGminimize_target, x0=[H_0, G_0], bounds=(HRange, GRange))

        Herr = 0
        Gerr = 0

        if self.errors is not None:
            Hs = []
            Gs = []

            for _ in range(errorSimulation):

                e = np.random.normal(0, self.errors)

                runres = minimize(
                    self._HGminimize_target,
                    x0=[H_0, G_0],
                    bounds=(HRange, GRange),
                    args=(e,),
                )
                Hs.append(runres.x[0])
                Gs.append(runres.x[1])

            Herr = np.std(Hs)
            Gerr = np.std(Gs)

        return HGResult(res.x[0], res.x[1], Herr, Gerr, res.success, res.x[1] in GRange, len(self.mags))

    def _HGminimize_target(
        self, _hg: Tuple[float, float], error: Union[np.ndarray, float] = 0
    ) -> float:

        adj_mag = calcPhaseCurve(_hg[0], _hg[1], self.phaseAngle)
        return np.sqrt(
            np.sum(np.power(self.reducedMagHG + error - adj_mag, 2))
            / (len(self.reducedMagHG) - 1)
        )



cimport numpy as np
import numpy as np
cimport cython


ctypedef fused valid_arg:
    np.ndarray
    cython.float
    cython.double
    

cdef extern from "constants.h":
    cdef cython.float HG_SIMPLE_A1
    cdef cython.float HG_SIMPLE_A2 
    cdef cython.float HG_SIMPLE_B1 
    cdef cython.float HG_SIMPLE_B2 
    cdef cython.float HG_A1
    cdef cython.float HG_A2
    cdef cython.float HG_B1
    cdef cython.float HG_B2
    cdef cython.float HG_C1
    cdef cython.float HG_C2

cdef _calcPhis_simple(valid_arg phaseAngle, cython.float G):
    cdef valid_arg phi1, phi2, tan_val

    tan_val = np.tan(.5 * (phaseAngle * (np.pi/180)))
    phi1 = np.exp(-HG_SIMPLE_A1 * np.power(tan_val, HG_SIMPLE_B1))
    phi2 = np.exp(-HG_SIMPLE_A2 * np.power(tan_val, HG_SIMPLE_B2))
    return phi1, phi2

cdef _phaseIntegral_simple(valid_arg phaseAngle, cython.float G):
    cdef valid_arg phi1, phi2
    phi1, phi2 = _calcPhis_simple(phaseAngle, G)
    return (1-G) * phi1 + G * phi2

cdef _calcPhis(valid_arg phaseAngle, cython.float G):
    cdef valid_arg W, phi1, phi2, phiS, phiL, sin_pa, tan_val

    sin_pa = np.sin(phaseAngle* (np.pi/180))
    tan_val = np.tan(.5 * (phaseAngle * (np.pi/180)))

    W = np.exp(-90.56 * tan_val*tan_val)

    phiS = 1 - ((HG_C1*sin_pa)/((.119+1.341*sin_pa) - (.754*sin_pa*sin_pa)))
    phiL = np.exp(-HG_A1 * np.power(tan_val, HG_B1))
    phi1 = W*phiS + (1-W)*phiL

    phiS = 1 - ((HG_C2*sin_pa)/((.119+1.341*sin_pa) - (.754*sin_pa*sin_pa)))
    phiL = np.exp(-HG_A2 * np.power(tan_val, HG_B2))
    phi2 = W*phiS + (1-W)*phiL

    return phi1, phi2

cdef _phaseIntegral(valid_arg phaseAngle, cython.float G):
    
    cdef valid_arg phi1, phi2

    phi1, phi2 = _calcPhis(phaseAngle, G)
    return (1-G) * phi1 + G * phi2

cdef _HG(valid_arg apmag, valid_arg helioDist, valid_arg obsDist, valid_arg phaseAngle ,cython.float G, const cython.int simple):
    cdef valid_arg phaseInt
    if simple:
        phaseInt = _phaseIntegral_simple(phaseAngle, G)
    else:
        phaseInt = _phaseIntegral(phaseAngle, G)

    return apmag - (5 * np.log10(helioDist*obsDist) - 2.5*np.log10(phaseInt))


cdef _calcPhaseCurve(float H, float G, valid_arg phaseAngle):
    cdef float a1, a2, phi1, phi2

    a1 = np.power(10, -H/2.5) / (1+(G/(1-G)))
    a2 = G*a1/(1-G)

    phi1, phi2 = _calcPhis(phaseAngle, G)
    return -2.5 * np.log10(a1*phi1 + a2*phi2)

def calcPhaseCurve(float H, float G, valid_arg phaseAngle):
    r"""Calculate the reduced magnitudes, the Bowell HG system\ :footcite:p:`bowellHG` for the given phase angles.

    Will use the non-simple :math:`\phi_1,\phi_2` calculations

    :param H: The Bowell H value
    :type H: float
    :param G: The Bowell G value 
    :type G: float
    :param phaseAngle: The phase angle, the sun - object - observer angle (degrees)
    :type phaseAngle: np.ndarray
    :return: The reduced magnitudes
    :rtype: valid_arg
    """

    return _calcPhaseCurve(H, G, phaseAngle)

def bowellCalcAbsMag(valid_arg apmag, valid_arg helioDist, valid_arg obsDist, valid_arg phaseAngle ,const cython.float G, const cython.int simple = True):
    r"""Calculate the absolute mags for an object from the Bowell HG system\ :footcite:p:`bowellHG`
    
    

    :param apmag: The apparent magnitudes to use
    :type apmag: np.ndarray
    :param helioDist: The distance from the object to the Sun (AU)
    :type helioDist: np.ndarray
    :param obsDist: The distance from the object to the Observer/Earth (AU)
    :type obsDist: np.ndarray
    :param phaseAngle: The phase angle, the sun - object - observer angle (degrees)
    :type phaseAngle: np.ndarray
    :param G: The Bowell G value
    :type G: float
    :param simple: If to use the simple or other version of :math:`\phi` calculations. The usual version used is the simple version. defaults to True
    :type simple: bool
    :return: The absolute magnitudes for the observations.
    :rtype: np.ndarray
    """
    return _HG(apmag, helioDist, obsDist, phaseAngle ,G, simple)




cimport numpy as np
import numpy as np
cimport cython

ctypedef fused valid_arg:
    np.ndarray
    cython.float
    cython.double

cdef _calcReducedMag(valid_arg mag, valid_arg helioDist, valid_arg obsDist):
    return mag - 5*np.log10(helioDist*obsDist)

def calcReducedMag(valid_arg mag, valid_arg helioDist, valid_arg obsDist):
    r"""Calculate teh reduced magnitude
    ..math:

        V(\alpha) = m - 5\log10(\Delta r)
    
    :param mag: The magnitudes of the observations
    :type mag: np.ndarray
    :param helioDist: The heliocentric distances
    :type helioDist: np.ndarray
    :param obsDist: The observer/Earth distances
    :type obsDist: np.ndarray
    :return: The reduced magnitudes
    :rtype: np.ndarray
    """
    return _calcReducedMag(mag, helioDist, obsDist)


from ast import Call

import numpy as np
from astroquery.jplhorizons import HorizonsClass
from astropy import units as u
from typing import Callable
from .hg import bowellCalcAbsMag

class HorAbsMags(HorizonsClass):
    """Wrapper for astroquries Horizons class. can calculate the absolute magnitude.

    :param mags: The mags of observations at the provided ephocs
    :type mags: np.ndarray
    """

    def __init__(self, mags:np.ndarray, *args, **kwargs) -> None:
        """Constructor
        """
        self.mags:np.ndarray = mags
        eph = self.ephemerides()
        super().__init__(*args, **kwargs)

    def calcAbsMag(self, absMagCalcFunc:Callable = bowellCalcAbsMag, G:float=None) -> np.ndarray:
        """calcAbsMag Function to calculate the absolute magnitudes. Will grab the data from horizons

        :param absMagCalcFunc: The function that calcualtes the absolute magnitude, defaults to HG
        :type absMagCalcFunc: Callable, optional
        :param G: The G value to use. Use None if using the value from horizons, defaults to None
        :type G: float, optional
        :return: The absolute magnitudes for the observations.
        :rtype: np.ndarray
        """
        if G is None:
            G = self.eph["G"][0]
        helioDist = self.eph["r"]
        obsDist = self.eph["delta"]
        phaseAngle = self.eph["alpha"]

        return absMagCalcFunc(self.mags, helioDist, obsDist, phaseAngle, G)
    
    def getLightTimeCorrection(self, outputUnit:u = u.day)-> np.ndarray:
        """getLightTimeCorrection Calculate the light time corrction for the observations

        :param outputUnit: The output unit to use, defaults to u.day
        :type outputUnit: u, optional
        :return: The light time correction for each ephoch
        :rtype: np.ndarray
        """
        return self.eph["lighttime"].to(outputUnit).data


#!/usr/bin/env python

"""The setup script."""

import os
from setuptools import setup, find_packages, dist

dist.Distribution().fetch_build_eggs(["Cython>=0.29.0", "numpy>=1.20.0"])

from Cython.Build import cythonize
import numpy as np

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "astroquery>=0.4.1",
    "astropy>=4.0.0",
    "cython>=0.29.0",
    "numpy>=1.20.0",
    "scipy",
]

test_requirements = ["pytest>=3", "numpy>=1.20.0"]

header_files = ["phase_curve_tools/constants.h"]
cython_files = ["phase_curve_tools/*.pyx"]

modules = cythonize(cython_files, language_level="3")

setup(
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        "setuptools>=18.0",
        "cython",
        "numpy",
    ],
    author="Daniel Kramer",
    author_email="drk98@nau.edu",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="A Python package for asteroid phase curves",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="phase_curve_tools",
    name="phase_curve_tools",
    packages=find_packages(include=["phase_curve_tools", "phase_curve_tools.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/drk98/phase_curve_tools",
    version='0.4.0',
    zip_safe=False,
    ext_modules=modules,
    include_dirs=[np.get_include()],
    extra_compile_args=["-O3"],
)
